# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/3/5 01:04
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""
import numpy as np
import polars as pl

over = dict(
    partition_by=["date", "asset"],
    order_by=["time"]
)


def itd_mean(expr: pl.Expr, windows): return expr.rolling_mean(windows, min_samples=1).over(**over)


def itd_std(expr: pl.Expr, windows): return expr.rolling_std(windows, min_samples=1).over(**over)


def itd_sum(expr: pl.Expr, windows): return expr.rolling_sum(windows, min_samples=1).over(**over)


def itd_var(expr: pl.Expr, windows): return expr.rolling_var(windows, min_samples=1).over(**over)


def itd_skew(expr: pl.Expr, windows): return expr.rolling_skew(windows, ).over(**over)


def itd_ref(expr: pl.Expr, windows, dims):  # return expr.shift(int(abs(windows))).over(**over)
    return (
        expr
        .map_batches(
            lambda x: pl.DataFrame(
                x
                .to_numpy()
                .reshape(dims)
                .transpose((1, 0, 2))
                .reshape((dims[1], -1))
            )
            .shift(windows)
            .to_numpy()
            .reshape((dims[1], dims[0], dims[2]))
            .transpose((1, 0, 2))
            .ravel(),
            # return_dtype=pl.self_dtype()
        )
        .replace(np.nan, None)
    )


def itd_mid(expr: pl.Expr, windows): return expr.rolling_median(windows, min_samples=1).over(**over)


def itd_mad(expr: pl.Expr, windows):
    return 1.4826 * (expr - expr.rolling_median(windows, min_samples=1)).abs().rolling_median(windows, min_samples=1).over(**over)


def itd_rank(expr: pl.Expr, windows, dims):
    return (
        expr
        .map_batches(
            lambda x: pl.DataFrame(
                x
                .to_numpy()
                .reshape(dims)
                .transpose((1, 0, 2))
                .reshape((dims[1], -1))
            )
            .with_row_index()
            .rolling("index", period=f"{windows}i")
            .agg(pl.all().exclude("index").rank().last())
            .drop("index")
            .to_numpy()
            .reshape((dims[1], dims[0], dims[2]))
            .transpose((1, 0, 2))
            .ravel(),
            # return_dtype=pl.self_dtype()
        )
    )

def itd_prod(expr: pl.Expr, windows, dims):
    return (
        expr
        .map_batches(
            lambda x: pl.DataFrame(
                x
                .to_numpy()
                .reshape(dims)
                .transpose((1, 0, 2))
                .reshape((dims[1], -1))
            )
            .with_row_index()
            .rolling("index", period=f"{windows}i")
            .agg(pl.all().exclude("index").cum_prod())
            .drop("index")
            .to_numpy()
            .reshape((dims[1], dims[0], dims[2]))
            .transpose((1, 0, 2))
            .ravel(),
            # return_dtype=pl.self_dtype()
        )
    )

def itd_max(expr: pl.Expr, windows): return expr.rolling_max(windows, min_samples=1).over(**over)


def itd_min(expr: pl.Expr, windows): return expr.rolling_min(windows, min_samples=1).over(**over)


def itd_ewmmean(expr: pl.Expr, com=None, span=None, half_life=None, alpha=None):
    return (expr
            .ewm_mean(com=com,
                      span=span,
                      half_life=half_life,
                      alpha=alpha,
                      adjust=False,
                      min_samples=1)
            .over(**over))


def itd_ewmstd(expr: pl.Expr, com=None, span=None, half_life=None, alpha=None):
    return (expr
            .ewm_std(com=com,
                     span=span,
                     half_life=half_life,
                     alpha=alpha,
                     adjust=False,
                     min_samples=1)
            .over(**over))


def itd_ewmvar(expr: pl.Expr, com=None, span=None, half_life=None, alpha=None):
    return (expr
            .ewm_var(com=com,
                     span=span,
                     half_life=half_life,
                     alpha=alpha,
                     adjust=False,
                     min_samples=1)
            .over(**over))


def itd_cv(expr: pl.Expr, windows): return itd_std(expr, windows) / itd_mean(expr, windows)


def itd_snr(expr: pl.Expr, windows): return itd_mean(expr, windows) / itd_std(expr,
                                                                              windows)  # 信噪比: signal_to_noise ratio


def itd_diff(expr: pl.Expr, windows=1): return expr.diff(windows).over(**over)


def itd_pct(expr: pl.Expr, windows=1): return expr.pct_change(windows).over(**over)


def itd_corr(left: pl.Expr, right: pl.Expr, windows): return pl.rolling_corr(left, right, window_size=windows,
                                                                             min_samples=1).over(**over)


def itd_cov(left: pl.Expr, right: pl.Expr, windows): return pl.rolling_cov(left, right, window_size=windows,
                                                                           min_samples=1).over(**over).replace(np.nan,
                                                                                                               None)

def itd_slope(left: pl.Expr, right: pl.Expr, windows): return (
            itd_mean(left * right, windows) - itd_mean(right, windows) * itd_mean(left, windows)) / itd_var(right,
                                                                                                            windows)


def itd_resid(left: pl.Expr, right: pl.Expr, windows): return right - itd_slope(left, right, windows) * right


def itd_quantile(expr: pl.Expr, windows, quantile):
    return expr.rolling_quantile(window_size=windows, quantile=quantile, min_samples=1).over(**over)

def itd_entropy(expr: pl.Expr, windows, dims):
    return (
        expr
        .map_batches(
            lambda x: pl.DataFrame(
                x
                .to_numpy()
                .reshape(dims)
                .transpose((1, 0, 2))
                .reshape((dims[1], -1))
            )
            .with_row_index()
            .rolling("index", period=f"{windows}i")
            .agg(pl.all().exclude("index").entropy())
            .drop("index")
            .to_numpy()
            .reshape((dims[1], dims[0], dims[2]))
            .transpose((1, 0, 2))
            .ravel(),
            # return_dtype=pl.self_dtype()
        )
    )

def itd_zscore(expr: pl.Expr, windows):
    return (expr - itd_mean(expr, windows))/itd_std(expr, windows)

def itd_norm(expr: pl.Expr, windows):
    return (expr - itd_mid(expr, windows))/itd_mad(expr, windows)

def itd_fill_forward(expr: pl.Expr):
    return expr.fill_null(strategy="forward").over(**over)


# Copyright (c) ZhangYundi.
# Licensed under the MIT License. 
# Created on 2025/10/11 11:01
# Description:

import queue
import time
from collections.abc import Callable

import polars as pl
import threading
import logair

class DataService:

    def __init__(self, cache_size: int = 5):
        self._max_cache_size = cache_size
        self._cache = queue.Queue(maxsize=self._max_cache_size)
        self._cache_dict: dict[str, dict[str, pl.DataFrame]] = dict()  # 用于快速查找的字典
        self.stop_event = threading.Event()
        self._data_thread = None
        self.is_running = False
        self._fn = None

    def put_data(self, key: str, data: dict[str, pl.DataFrame]):
        self._cache.put(key)
        self._cache_dict[key] = data

    def get_data(self) -> pl.DataFrame:
        try:
            key = self._cache.get_nowait()
            data = self._cache_dict.pop(key)
            return key, data, False
        except queue.Empty:
            return "", None, True

    def _data_loading_worder(self,
                             keys: list[str],
                             iter_conf: dict[str, list[str]],):
        logger = logair.get_logger(f"{__name__}.{self.__class__.__name__}.worker")
        logger.info(f"Data loading worker started for {len(keys)} keys.")

        def worker(key, work_id: int):
            result = dict()
            try:
                for name, iters in iter_conf.items():
                    data = self._fn(key=key, iterables=iters)
                    result[name] = data
                self.put_data(key, result)
                logger.info(f"{key}(WorkerID: {work_id}) Loaded data.")
            except Exception as e:
                logger.warning(f"Failed to load data for {key}(WorkerID: {work_id}): {e}")

        for i, k in enumerate(keys):
            worker(key=k, work_id=i + 1)
        self.stop_event.set()

    def start(self,
              fn: Callable,
              keys: list[str],
              iter_conf: dict[str, list[str]],
              max_cache_size: int,):
        """

        Parameters
        ----------
        fn: 获取数据的函数，参数为 key 和 iterables 以及其它参数
        keys
        iter_conf
        max_cache_size

        Returns
        -------

        """
        logger = logair.get_logger(f"{__name__}.{self.__class__.__name__}")
        self._fn = fn
        self._max_cache_size = max_cache_size
        # 先确保之前的服务已经完全停止
        if self.is_running:
            logger.warning("DataService is already running")
            self.stop()
            # return
        # 重新初始化缓存和 stop_event
        self._cache = queue.Queue(maxsize=self._max_cache_size)
        self._cache_dict.clear()
        self.stop_event.clear()

        logger.info(f"Starting DataService({self._max_cache_size}) for {len(keys)} key...")
        # 启动后台数据加载线程
        self._data_thread = threading.Thread(
            target=self._data_loading_worder,
            args=(keys,
                  iter_conf,),
            daemon=True,  # 设置为守护线程，主程序退出时自动结束
        )
        self.is_running = True
        self._data_thread.start()
        logger.info("DataService started successfully.")

    def stop(self):
        """停止数据服务"""
        logger = logair.get_logger(f"{__name__}.{self.__class__.__name__}")
        if not self.is_running:
            logger.warning("Data service is not running")
            return
        logger.info("Stopping data service...")
        self.stop_event.set()
        if self._data_thread and self._data_thread.is_alive():
            self._data_thread.join(timeout=10)
        self.is_running = False
        logger.info("Data service stopped")

    def do(self, consumer: callable, wait_secs: float = 3):
        """
        消费数据
        Parameters
        ----------
        consumer:
        wait_secs

        Returns
        -------

        """
        while self.is_running:
            key, data, is_empty = self.get_data()
            if is_empty:
                if self.stop_event.is_set():
                    self.stop()
                    break
                else:
                    time.sleep(wait_secs)
                    continue
            consumer(dict(key=key, data=data))


D = DataService()


# Copyright (c) ZhangYundi.
# Licensed under the MIT License. 
# Created on 2025/7/17 14:40
# Description:

from pathlib import Path
from dynaconf import Dynaconf
import logair


USERHOME = Path("~").expanduser() # 用户家目录
NAME = "lidb"
DB_PATH = USERHOME / NAME
CONFIG_PATH = USERHOME / ".config" / NAME / "settings.toml"

logger = logair.get_logger(NAME)


if not CONFIG_PATH.exists():
    try:
        CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        logger.error(f"Failed to create settings file: {e}")
    with open(CONFIG_PATH, "w") as f:
        template_content = f'[global]\npath="{DB_PATH}"\n'
    with open(CONFIG_PATH, "w") as f:
        f.write(template_content)
    logger.info(f"Settings file created: {CONFIG_PATH}")

def get_settings():
    try:
        return Dynaconf(settings_files=[CONFIG_PATH])
    except Exception as e:
        logger.error(f"Read settings file failed: {e}")
        return {}

# 读取配置文件覆盖
_settiings = get_settings()
if _settiings is not None:
    setting_db_path = _settiings.get(f"global.path", "")
    if setting_db_path:
        DB_PATH = Path(setting_db_path)



# Copyright (c) ZhangYundi.
# Licensed under the MIT License. 

from .init import (
    NAME,
    DB_PATH,
    CONFIG_PATH,
    get_settings,
)

from .database import (
    sql,
    put,
    has,
    tb_path,
    read_mysql,
    write_mysql,
    execute_mysql,
    read_ck,
    scan,
)

from .table import Table, TableMode
from .dataset import Dataset, DataLoader
from .qdf import to_lazy, Expr
from .svc import DataService, D

from .parse import parse_hive_partition_structure

__version__ = "1.1.19"


# Copyright (c) ZhangYundi.
# Licensed under the MIT License. 
# Created on 2025/10/27 14:13
# Description:

from __future__ import annotations

from collections import defaultdict
from enum import Enum
from functools import partial
from typing import Callable, Literal

import logair
import polars as pl
import polars.selectors as cs
import xcals
import ygo

from .database import put, tb_path, scan, DB_PATH
from .parse import parse_hive_partition_structure
from .qdf import to_lazy, LQDF


class InstrumentType(Enum):
    STOCK = "Stock"  # 股票
    ETF = "ETF"  #
    CB = "ConvertibleBond"  # 可转债


def complete_data(fn, date, save_path, partitions):
    logger = logair.get_logger(__name__)
    try:
        data = fn(date=date)
        if data is None:
            # 保存数据的逻辑在fn中实现了
            return
        # 剔除以 `_` 开头的列
        data = data.select(~cs.starts_with("_"))
        if not isinstance(data, (pl.DataFrame, pl.LazyFrame)):
            logger.error(f"{save_path}: Result of dataset.fn must be polars.DataFrame or polars.LazyFrame.")
            return
        if isinstance(data, pl.LazyFrame):
            data = data.collect()
        cols = data.columns
        if "date" not in cols:
            data = data.with_columns(pl.lit(date).alias("date")).select("date", *cols)

        put(data, save_path, partitions=partitions)
    except Exception as e:
        logger.error(f"{save_path}: Error when complete data for {date}")
        logger.warning(e)


class Dataset:

    def __init__(self,
                 fn: Callable[..., pl.DataFrame],
                 tb: str,
                 update_time: str = "",
                 partitions: list[str] = None,
                 by_asset: bool = True,
                 by_time: bool = False):
        """

        Parameters
        ----------
        fn: str
            数据集计算函数
        tb: str
            数据集保存表格
        update_time: str
            更新时间: 默认没有-实时更新，也就是可以取到当天值
        partitions: list[str]
            分区
        by_asset: bool
            是否按照标的进行分区，默认 True
        by_time: bool
            是否按照标的进行分区，默认 False
        """
        self.fn = fn
        self.fn_params_sig = ygo.fn_signature_params(fn)
        self._by_asset = by_asset
        self._by_time = by_time
        self._append_partitions = ["asset", "date"] if by_asset else ["date", ]
        if by_time:
            self._append_partitions.append("time")
        if partitions is not None:
            partitions = [k for k in partitions if k not in self._append_partitions]
            partitions = [*partitions, *self._append_partitions]
        else:
            partitions = self._append_partitions
        self.partitions = partitions
        self._type_asset = "asset" in self.fn_params_sig
        self.update_time = update_time

        self.tb = tb
        self.save_path = tb_path(tb)
        fn_params = ygo.fn_params(self.fn)
        self.fn_params = {k: v for (k, v) in fn_params}
        self.constraints = dict()
        for k in self.partitions[:-len(self._append_partitions)]:
            if k in self.fn_params:
                v = self.fn_params[k]
                if isinstance(v, (list, tuple)) and not isinstance(v, str):
                    v = sorted(v)
                self.constraints[k] = v
                self.save_path = self.save_path / f"{k}={v}"

    def is_empty(self, path) -> bool:
        return not any(path.rglob("*.parquet"))

    def __call__(self, *fn_args, **fn_kwargs):
        # self.fn =
        fn = partial(self.fn, *fn_args, **fn_kwargs)
        ds = Dataset(fn=fn,
                     tb=self.tb,
                     partitions=self.partitions,
                     by_asset=self._by_asset,
                     by_time=self._by_time,
                     update_time=self.update_time)
        return ds

    def get_value(self, date, eager: bool = True, **constraints):
        """
        取值: 不保证未来数据
        Parameters
        ----------
        date: str
            取值日期
        eager: bool
        constraints: dict
            取值的过滤条件

        Returns
        -------

        """
        _constraints = {k: v for k, v in constraints.items() if k in self.partitions}
        _limits = {k: v for k, v in constraints.items() if k not in self.partitions}
        search_path = self.save_path
        for k, v in _constraints.items():
            if isinstance(v, (list, tuple)) and not isinstance(v, str):
                v = sorted(v)
            search_path = search_path / f"{k}={v}"
        search_path = search_path / f"date={date}"

        if not self.is_empty(search_path):
            lf = scan(search_path).cast({"date": pl.Utf8})
            schema = lf.collect_schema()
            _limits = {k: v for k, v in constraints.items() if schema.get(k) is not None}
            lf = lf.filter(date=date, **_limits)
            if not eager:
                return lf
            data = lf.collect()
            if not data.is_empty():
                return data
        fn = self.fn
        save_path = self.save_path

        if self._type_asset:
            if "asset" in _constraints:
                fn = ygo.delay(self.fn)(asset=_constraints["asset"])
        if len(self.constraints) < len(self.partitions) - len(self._append_partitions):
            # 如果分区指定的字段没有在Dataset定义中指定，需要在get_value中指定
            params = dict()
            for k in self.partitions[:-len(self._append_partitions)]:
                if k not in self.constraints:
                    v = constraints[k]
                    params[k] = v
                    save_path = save_path / f"{k}={v}"
            fn = ygo.delay(self.fn)(**params)
        logger = logair.get_logger(__name__)

        today = xcals.today()
        now = xcals.now()
        if (date > today) or (date == today and now < self.update_time):
            logger.warning(f"{self.tb}: {date} is not ready, waiting for {self.update_time}")
            return
        complete_data(fn, date, save_path, self._append_partitions)

        lf = scan(search_path).cast({"date": pl.Utf8})
        schema = lf.collect_schema()
        _limits = {k: v for k, v in constraints.items() if schema.get(k) is not None}
        lf = lf.filter(date=date, **_limits)
        if not eager:
            return lf
        return lf.collect()

    def get_pit(self, date: str, query_time: str, eager: bool = True, **contraints):
        """取值：如果取值时间早于更新时间，则返回上一天的值"""
        if not self.update_time:
            return self.get_value(date, **contraints)
        val_date = date
        if query_time < self.update_time:
            val_date = xcals.shift_tradeday(date, -1)
        return self.get_value(val_date, eager=eager, **contraints).with_columns(date=pl.lit(date), )

    def get_history(self,
                    dateList: list[str],
                    n_jobs: int = 5,
                    backend: Literal["threading", "multiprocessing", "loky"] = "loky",
                    eager: bool = True,
                    rep_asset: str = "000001",  # 默认 000001
                    **constraints):
        """获取历史值: 不保证未来数据"""
        _constraints = {k: v for k, v in constraints.items() if k in self.partitions}
        search_path = self.save_path
        for k, v in _constraints.items():
            if isinstance(v, (list, tuple)) and not isinstance(v, str):
                v = sorted(v)
            search_path = search_path / f"{k}={v}"
        if self.is_empty(search_path):
            # 需要补全全部数据
            missing_dates = dateList
        else:
            if not self._type_asset:
                _search_path = self.save_path
                for k, v in _constraints.items():
                    if k != "asset":
                        _search_path = _search_path / f"{k}={v}"
                    else:
                        _search_path = _search_path / f"asset={rep_asset}"
                hive_info = parse_hive_partition_structure(_search_path)
            else:
                hive_info = parse_hive_partition_structure(search_path)
            exist_dates = hive_info["date"].to_list()
            missing_dates = set(dateList).difference(set(exist_dates))
            missing_dates = sorted(list(missing_dates))
        if missing_dates:
            fn = self.fn
            save_path = self.save_path

            if self._type_asset:
                if "asset" in _constraints:
                    fn = ygo.delay(self.fn)(asset=_constraints["asset"])

            if len(self.constraints) < len(self.partitions) - len(self._append_partitions):
                params = dict()
                for k in self.partitions[:-len(self._append_partitions)]:
                    if k not in self.constraints:
                        v = constraints[k]
                        params[k] = v
                        save_path = save_path / f"{k}={v}"
                fn = ygo.delay(self.fn)(**params)

            with ygo.pool(n_jobs=n_jobs, backend=backend) as go:
                info_path = self.save_path
                try:
                    info_path = info_path.relative_to(DB_PATH)
                except:
                    pass
                for date in missing_dates:
                    go.submit(complete_data, job_name=f"Completing {info_path}")(
                        fn=fn,
                        date=date,
                        save_path=save_path,
                        partitions=self._append_partitions,
                    )
                go.do()
        data = scan(search_path, ).cast({"date": pl.Utf8}).filter(pl.col("date").is_in(dateList), **constraints)
        data = data.sort("date")
        if eager:
            return data.collect()
        return data


def loader(data_name: str,
           ds: Dataset,
           date_list: list[str],
           prev_date_list: list[str],
           prev_date_mapping: dict[str, str],
           time: str,
           **constraints) -> pl.LazyFrame:
    if time < ds.update_time:
        if len(prev_date_list) > 1:
            lf = ds.get_history(prev_date_list, eager=False, **constraints)
        else:
            lf = ds.get_value(prev_date_list[0], eager=False, **constraints)
    else:
        if len(date_list) > 1:
            lf = ds.get_history(date_list, eager=False, **constraints)
        else:
            lf = ds.get_value(date_list[0], eager=False, **constraints)
    schema = lf.collect_schema()
    include_time = schema.get("time") is not None
    if include_time:
        lf = lf.filter(time=time)
    else:
        lf = lf.with_columns(time=pl.lit(time))
    if time < ds.update_time:
        lf = lf.with_columns(date=pl.col("date").replace(prev_date_mapping))
    return data_name, lf


def load_ds(ds_conf: dict[str, list[Dataset]],
            beg_date: str,
            end_date: str,
            time: str,
            n_jobs: int = 7,
            backend: Literal["threading", "multiprocessing", "loky"] = "threading",
            eager: bool = False,
            **constraints) -> dict[str, pl.DataFrame|pl.LazyFrame]:
    """
    加载数据集
    Parameters
    ----------
    ds_conf: dict[str, list[Dataset]]
        数据集配置: key-data_name, value-list[Dataset]
    beg_date: str
        开始日期
    end_date: str
        结束日期
    time: str
        取值时间
    n_jobs: int
        并发数量
    backend: str
    eager: bool
        是否返回 DataFrame
        - True: 返回DataFrame
        - False: 返回LazyFrame
    constraints
        限制条件，比如 asset='000001'
    Returns
    -------
    dict[str, polars.DataFrame | polars.LazyFrame]
        - key: data_name
        - value: polars.DataFrame

    """
    if beg_date > end_date:
        raise ValueError("beg_date must be less than end_date")
    date_list = xcals.get_tradingdays(beg_date, end_date)
    beg_date, end_date = date_list[0], date_list[-1]
    prev_date_list = xcals.get_tradingdays(xcals.shift_tradeday(beg_date, -1), xcals.shift_tradeday(end_date, -1))
    prev_date_mapping = {prev_date: date_list[i] for i, prev_date in enumerate(prev_date_list)}
    results = defaultdict(list)
    with ygo.pool(n_jobs=n_jobs, backend=backend) as go:
        for data_name, ds_list in ds_conf.items():
            for ds in ds_list:
                go.submit(loader,
                          job_name="Loading",
                          postfix=data_name)(data_name=data_name,
                                             ds=ds,
                                             date_list=date_list,
                                             prev_date_list=prev_date_list,
                                             prev_date_mapping=prev_date_mapping,
                                             time=time,
                                             **constraints)
        for name, lf in go.do():
            results[name].append(lf)
    index = ("date", "time", "asset")
    LFs = {
        name: (pl.concat(lfList, how="align")
               .sort(index)
               .select(*index,
                       cs.exclude(index))
               )
        for name, lfList in results.items()}
    if not eager:
        return LFs
    return {
        name: lf.collect()
        for name, lf in LFs.items()
    }

class DataLoader:

    def __init__(self, name: str):
        self._name = name
        self._lf: pl.LazyFrame = None
        self._df: pl.DataFrame = None
        self._index: tuple[str] = ("date", "time", "asset")
        self._db: LQDF = None
        self._one: pl.DataFrame = None

    def get(self,
            ds_list: list[Dataset],
            beg_date: str,
            end_date: str,
            n_jobs: int = 11,
            backend: Literal["threading", "multiprocessing", "loky"] = "threading",
            **constraints):
        """
        添加数据集
        Parameters
        ----------
        data_name: str
        ds_list: list[Dataset]
        beg_date: str
        end_date: str
        n_jobs: int
        backend: str
        constraints

        Returns
        -------

        """
        lf = load_ds(ds_conf={self._name: ds_list},
                     beg_date=beg_date,
                     end_date=end_date,
                     n_jobs=n_jobs,
                     backend=backend,
                     eager=False,
                     **constraints)
        self._lf = lf.get(self._name)
        self._df = None
        self._db = to_lazy(self._lf, self._index, align=True)
        dateList = xcals.get_tradingdays(beg_date, end_date)
        _data_name = f"{self._name}(one_day)"
        self._one = load_ds(ds_conf={_data_name: ds_list},
                            beg_date=dateList[0],
                            end_date=dateList[0],
                            n_jobs=n_jobs,
                            backend=backend,
                            eager=False,
                            **constraints).get(_data_name).collect()

    @property
    def name(self) -> str:
        return self._name

    @property
    def one_day(self) -> pl.DataFrame:
        return self._one

    @property
    def schema(self) -> pl.Schema:
        return self._one.schema

    @property
    def columns(self) -> list[str]:
        return self._one.columns

    def collect(self) -> pl.DataFrame:
        if self._df is None:
            self._df = self._lf.collect()
        return self._df

    def select(self, *exprs: str) -> pl.DataFrame:
        return self._db.sql(*exprs).collect()




# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/3/5 21:40
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""
from __future__ import annotations

import importlib.util
import sys
from functools import lru_cache
from functools import partial
from pathlib import Path

import polars as pl
import inspect
import logair

from .errors import CalculateError, CompileError, PolarsError, FailError
from .expr import Expr

# 动态加载模块
module_name = "udf"
module_path = Path(__file__).parent / "udf" / "__init__.py"
spec = importlib.util.spec_from_file_location(module_name, module_path)
module = importlib.util.module_from_spec(spec)
sys.modules[module_name] = module
spec.loader.exec_module(module)

logger = logair.get_logger("quda.qdf")


@lru_cache(maxsize=512)
def parse_expr(expr: str) -> Expr:
    return Expr(expr)


class QDF:

    def __init__(self,
                 data: pl.LazyFrame | pl.DataFrame,
                 index: tuple[str] = ("date", "time", "asset"),
                 align: bool = True, ):
        assert isinstance(data, (pl.LazyFrame, pl.DataFrame)), "data must be a polars DataFrame or LazyFrame"
        data = data.lazy()
        data = data.cast({pl.Decimal: pl.Float64})# .cast({pl.Float32: pl.Float64})
        self.data = data.collect()
        self.index = index
        self.dims = [self.data[name].drop_nulls().n_unique() for name in index]
        if align:
            lev_vals: list[pl.DataFrame] = [self.data.select(name).unique() for name in index]
            full_index = lev_vals[0]
            for lev_val in lev_vals[1:]:
                full_index = full_index.join(lev_val, how="cross")
            self.data = full_index.join(self.data, on=index, how='left')
        self.data = self.data.sort(self.index)
        self.failed = list()
        self._expr_cache = dict()  # type: dict[Expr, str]
        self._cur_expr_cache = dict()
        self._data_: pl.LazyFrame = None

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.data.__str__()

    def register_udf(self, func: callable, name: str = None):
        name = name if name is not None else func.__name__
        setattr(module, name, func)

    def _compile_expr(self, expr: str, cover: bool):
        """str表达式 -> polars 表达式"""
        try:
            expr_parsed = Expr(expr)
            alias = expr_parsed.alias  # if expr_parsed.alias is not None else str(expr_parsed)
            current_cols = set(self.data.columns)
            if alias in current_cols and not cover:
                return pl.col(alias), alias
            # 如果该表达式已有对应列，直接复用
            if expr_parsed in self._expr_cache and not cover:
                expr_pl: pl.Expr = pl.col(self._expr_cache[expr_parsed]).alias(alias)
                self._data_ = self._data_.with_columns(expr_pl)
                return pl.col(alias), alias
            elif expr_parsed in self._cur_expr_cache and not cover:
                expr_pl: pl.Expr = pl.col(self._cur_expr_cache[expr_parsed]).alias(alias)
                self._data_ = self._data_.with_columns(expr_pl)
                return pl.col(alias), alias

            def recur_compile(expr_: Expr):
                """递归编译"""
                alias_ = expr_.alias
                if alias_ in current_cols and not cover:
                    # 已存在：直接select数据源
                    return pl.col(alias_)
                if expr_ in self._expr_cache:
                    return pl.col(self._expr_cache[expr_]).alias(alias_)
                elif expr_ in self._cur_expr_cache:
                    return pl.col(self._cur_expr_cache[expr_]).alias(alias_)
                func = getattr(module, expr_.fn_name)
                _params = sorted(list(inspect.signature(func).parameters.keys()))
                if "dims" in _params:
                    func = partial(func, dims=self.dims)
                args = list()
                kwargs = dict()
                for arg in expr_.args:
                    if isinstance(arg, Expr):
                        args.append(recur_compile(arg))
                    elif isinstance(arg, dict):
                        kwargs.update(arg)
                    elif isinstance(arg, str):
                        if arg.lower() == "null":
                            args.append(None)
                        else:
                            args.append(pl.col(arg))
                    else:
                        args.append(arg)  # or args.append(pl.lit(arg))
                try:
                    expr_pl: pl.Expr = func(*args, **kwargs)
                    self._data_ = self._data_.with_columns(expr_pl.fill_nan(None).alias(alias_))
                    self._cur_expr_cache[expr_] = alias_
                    return pl.col(alias_)
                except Exception as e:
                    raise CompileError(message=f"{expr_.fn_name}({', '.join([str(arg) for arg in args])})\n{e}") from e

            return recur_compile(expr_parsed), alias
        except (CalculateError, CompileError, PolarsError) as e:
            raise e
        except Exception as e:
            # 所有未处理的错误统一抛出为 CompileError
            raise CompileError(message=f"[编译器外层]\n{e}") from e

    def sql(self, *exprs: str, cover: bool = False, ) -> pl.LazyFrame:
        """
        表达式查询
        Parameters
        ----------
        exprs: str
            表达式，比如 "ts_mean(close, 5) as close_ma5"
        cover: bool
            当遇到已经存在列名的时候，是否重新计算覆盖原来的列, 默认False，返回已经存在的列，跳过计算
            - True: 重新计算并且返回新的结果，覆盖掉原来的列
            - False, 返回已经存在的列，跳过计算
        Returns
        -------
            polars.DataFrame
        """
        self.failed = list()
        exprs_to_add = list()
        exprs_select = list()
        self._cur_expr_cache = {}
        self._data_ = self.data.lazy()

        for expr in exprs:
            try:
                compiled, alias = self._compile_expr(expr, cover)
                if compiled is not None:
                    exprs_to_add.append(compiled)
                    exprs_select.append(alias)
            except Exception as e:
                self.failed.append(FailError(expr, e))
        if self.failed:
            logger.warning(f"sql failed num：{len(self.failed)}/{len(exprs)}: \n {self.failed}")
        self._data_ = self._data_.fill_nan(None)
        new_expr_cache = dict()
        try:
            raw_cols = set(self.data.columns)
            current_cols = set(self._data_.collect_schema().names())
            raw_cols.update(set(exprs_select))
            # 缓存整理：只保留当前表达式的缓存
            self._expr_cache.update(self._cur_expr_cache)
            for k, v in self._expr_cache.items():
                if v in current_cols:
                    new_expr_cache[k] = v
            self._expr_cache = new_expr_cache
            drop_cols = current_cols.difference(raw_cols)
            self.data = self._data_.drop(drop_cols).collect()
            final_df = self.data.select(*self.index, *exprs_select)
            return final_df
        except Exception as e:
            # 缓存整理：只保留当前表达式的缓存
            resume_cols = set(self.data.columns)
            for k, v in self._expr_cache.items():
                if v in resume_cols:
                    new_expr_cache[k] = v
            self._expr_cache = new_expr_cache
            raise PolarsError(message=f"LazyFrame.collect() step error:\n{e}") from e



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/3/4 20:20
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

import polars as pl

over = dict(
    partition_by=["date", "time"],
    order_by=["asset", ]
)

EPS = 1e-12


def cs_ufit(expr: pl.Expr): return (expr - expr.median().over(**over)).abs()


def cs_rank(expr: pl.Expr): return expr.rank().over(**over)


def cs_demean(expr: pl.Expr): return expr - expr.mean().over(**over)


def cs_mean(expr: pl.Expr): return expr.mean().over(**over)


def cs_mid(expr: pl.Expr): return expr.median().over(**over)


def cs_moderate(expr: pl.Expr): return (expr - expr.mean().over(**over)).abs()


def cs_qcut(expr: pl.Expr, N=10):
    return expr.qcut(N, labels=[str(i) for i in range(1, N + 1)], allow_duplicates=True).cast(pl.Int32)


def cs_ic(left: pl.Expr, right: pl.Expr, ): return pl.corr(left, right, method="spearman").over(**over)


def cs_corr(left: pl.Expr, right: pl.Expr): return pl.corr(left, right, method="pearson").over(**over)


def cs_std(expr: pl.Expr): return expr.std().over(**over)


def cs_var(expr: pl.Expr): return expr.var().over(**over)


def cs_skew(expr: pl.Expr): return expr.skew().over(**over)


def cs_slope(left: pl.Expr, right: pl.Expr): return cs_corr(left, right) * cs_std(left) / cs_std(right)


def cs_resid(left: pl.Expr, right: pl.Expr): return left - cs_slope(left, right) * right


def cs_mad(expr: pl.Expr):
    return 1.4826 * (expr - expr.median()).abs().median().over(**over)


def cs_zscore(expr: pl.Expr): return (expr - cs_mean(expr)) / cs_std(expr)


def cs_norm(expr: pl.Expr): return (expr - cs_mid(expr)) / cs_mad(expr)


def cs_midby(expr: pl.Expr, *by: pl.Expr): return expr.median().over(partition_by=[*over.get("partition_by"), *by],
                                                                     order_by=over.get("order_by"))


def cs_madby(expr: pl.Expr, *by: pl.Expr): return 1.4826 * (expr - expr.median()).abs().median().over(
    partition_by=[*over.get("partition_by"), *by], order_by=over.get("order_by"))


def cs_normby(expr: pl.Expr, *by: pl.Expr): return (expr - cs_midby(expr, *by)) / (cs_madby(expr, *by) + EPS)


def cs_meanby(expr: pl.Expr, *by: pl.Expr): return expr.mean().over(partition_by=[*over.get("partition_by"), *by],
                                                                    order_by=over.get("order_by"))


def cs_stdby(expr: pl.Expr, *by: pl.Expr): return expr.std().over(partition_by=[*over.get("partition_by"), *by],
                                                                  order_by=over.get("order_by"))


def cs_sumby(expr: pl.Expr, *by: pl.Expr): return expr.sum().over(partition_by=[*over.get("partition_by"), *by],
                                                                  order_by=over.get("order_by"))


def cs_max(expr: pl.Expr): return expr.max().over(**over)


def cs_min(expr: pl.Expr): return expr.min().over(**over)


def cs_peakmax(expr: pl.Expr): return expr.peak_max().over(**over)


def cs_peakmin(expr: pl.Expr): return expr.peak_min().over(**over)


def cs_zscoreby(expr: pl.Expr, *by: pl.Expr): return (expr - cs_meanby(expr, *by)) / cs_stdby(expr, *by)


def cs_entropy(expr: pl.Expr): return expr.entropy().over(**over)


def cs_entropyby(expr: pl.Expr, *by: pl.Expr): return expr.entropy().over(partition_by=[*over.get("partition_by"), *by],
                                                                          order_by=over.get("order_by"))
